# import linkml
import logging
import random
import re
# urllib? urllib3? pure requests?
import urllib

import click
import click_log
import linkml.utils.rawloader as rl
import linkml_runtime
import pandas as pd
import requests
import yaml
from linkml_runtime.dumpers import yaml_dumper
# cosine? SIFT4?
from strsimpy.cosine import Cosine

# # for querying and changing?
# import linkml_runtime_api

# # PROGRESS
# # took out globals
# # reusing requests sessions
# # saving annotations
# # logging


# # TODO
# # add caching of already searched terms
# # add default values for functions

# todo this silences
#   SettingWithCopyWarning: A value is trying to be set on a copy of a slice from a DataFrame
#   but I should really be dealing with it
pd.options.mode.chained_assignment = None  # default='warn'

logger = logging.getLogger(__name__)
click_log.basic_config(logger)


# make classes out of DataFrames that would be appealing as globals
# TODO try dataclasses
class DataFrameClass:
    def __init__(self):
        self.mapping_frame = pd.DataFrame()

    def add(self, miniframe):
        self.mapping_frame = self.mapping_frame.append(miniframe)

    def get(self):
        return self.mapping_frame


def parse_yaml_file(yaml_file_name):
    with open(yaml_file_name, 'r') as stream:
        try:
            parse_res = yaml.safe_load(stream)
            return parse_res
        except yaml.YAMLError as exc:
            # log or print?
            logger.error(exc)


def dict_to_schema(dict_param):
    converted_schema = rl.load_raw_schema(dict_param)
    return converted_schema


def request_an_enum(schema_param, enum_name_param):
    enum_requested = schema_param.enums[enum_name_param]
    return enum_requested


def request_pvs(enum_param):
    pvs_requested = enum_param.permissible_values
    return pvs_requested


# make sorting optional?
def get_pv_names(pv_param):
    pv_names = [k for k, v in pv_param.items()]
    pv_names.sort()
    return pv_names


def make_cosine_obj(shingle_size_param):
    made_cosine_obj = Cosine(shingle_size_param)
    return made_cosine_obj


def ols_term_search(term, chars_to_whiteout, ontology_param, qf_param, rowcount_param, blank_row_param,
                    global_frame_param, session_param, ols_search_base_url, do_trunc):
    woed = do_whiteout(term, chars_to_whiteout, do_trunc)

    request_string = ols_search_base_url + \
                     '?q=' + \
                     urllib.parse.quote(woed) + '&' + \
                     'type=class' + '&' + \
                     'exact=false' + '&' + \
                     ontology_param + "&" + \
                     'rows=' + str(rowcount_param) + '&' + \
                     qf_param

    logger.debug(request_string)

    # this gets matching terms but doesn't show why they matched
    response_param = session_param.get(request_string)

    ols_string_search_res_j = response_param.json()
    ols_string_search_res_frame = pd.DataFrame(ols_string_search_res_j['response']['docs'])
    ols_string_search_res_frame.insert(0, "raw_query", term)
    ols_string_search_res_frame.insert(0, "tidied_query", woed)

    # did the string search get any result rows?
    r, c = ols_string_search_res_frame.shape
    if r == 0:
        no_search_res_dict = blank_row_param.copy()
        # no_search_res_dict['id'] = term
        no_search_res_dict['raw_query'] = term
        no_search_res_dict['tidied_query'] = woed
        no_search_res_frame = pd.DataFrame([no_search_res_dict])
        ols_string_search_res_frame = ols_string_search_res_frame.append(no_search_res_frame)
        # failures.append(orig_enum)

    global_frame_param.add(ols_string_search_res_frame)
    # return ols_string_search_res_frame?
    return True


# TODO refactor
def make_ontolgy_phrase(ontology_param):
    ontologies_phrase = ''
    if ontology_param is not None and ontology_param != "":
        ontologies_phrase = 'ontology=' + ontology_param.lower()
    return ontologies_phrase


# TODO refactor
def make_qf_phrase(qf_param):
    qf_phrase = ''
    if qf_param is not None and qf_param != "":
        qf_phrase = 'queryFields=' + qf_param.lower()
    return qf_phrase


def do_whiteout(raw_string, chars_to_whiteout, do_trunc):
    if chars_to_whiteout is not None and chars_to_whiteout != "":
        tidied_string = re.sub(r'[' + chars_to_whiteout + ']+', ' ', raw_string)
    else:
        tidied_string = raw_string
    if do_trunc:
        tidied_string = re.sub(r' \(.*$', '', tidied_string)
    return tidied_string


def get_ols_term_annotations(iri_param, ontology_param, session_param, ols_terms_based_url, term_annotations):
    logger.info(iri_param)
    once = urllib.parse.quote(iri_param, safe='')
    twice = urllib.parse.quote(once, safe='')
    # build url from base
    term_retr_assembled = ols_terms_based_url + ontology_param + '/terms/' + twice
    term_details = session_param.get(term_retr_assembled)
    term_json = term_details.json()

    if 'label' in set(term_json.keys()):
        # logger.debug(term_retr_assembled)
        # term_label = term_json['label']
        # logger.debug(term_label)
        label_frame = pd.DataFrame([[term_json['label'], 'label', 'label', '']],
                                   columns=['name', 'scope', 'type', 'xrefs'])
        label_frame['obo_id'] = term_json['obo_id']
        label_frame['pref_lab'] = term_json['label']

        label_frame.insert(0, "iri", iri_param)

        term_annotations.add(label_frame)

    if 'obo_synonym' in set(term_json.keys()):
        obo_syn_json = term_json['obo_synonym']
        obo_syn_frame = pd.DataFrame(obo_syn_json)
        obo_syn_frame['obo_id'] = term_json['obo_id']
        obo_syn_frame['pref_lab'] = term_json['label']
        obo_syn_frame.insert(0, "iri", iri_param)
        term_annotations.add(obo_syn_frame)

    return True


@click.command()
@click_log.simple_verbosity_option(logger)
@click.option('--modelfile', help="path to LinkML input", type=click.Path(exists=True), show_default=True,
              required=True)
# parametrize this so this script can be written many times an not overwrite the all mappings file
@click.option('--all_mappings_fn', default="target/all_mappings_frame.tsv",
              help="where do you want to write a table of all mappings?",
              type=click.Path(), show_default=True)
# binomial_name_enum
@click.option('--requested_enum_name', help="name of the enumeration that contains the terms you want to map",
              required=True)
@click.option('--overwrite_meaning/--no_overwrite', help="do you want to overwrite meanings that are already present?",
              default=True, show_default=True)
@click.option('--whiteout_chars', help="characters in terms that should be replaced with whitespace before mapping",
              default='._-', show_default=True)
# ontology_string = "NCBItaxon,PATO"
@click.option('--ontology_string', help="comma separated list of ontologies to use in the mapping", default='NCBItaxon',
              show_default=True)
@click.option('--ols_search_base_url', help="", default='http://www.ebi.ac.uk/ols/api/search', show_default=True)
@click.option('--ols_terms_based_url', help="", default='http://www.ebi.ac.uk/ols/api/ontologies/',
              show_default=True)
@click.option('--desired_row_count', help="how many rows of mappings do you want to retrieve for each term?", default=5,
              show_default=True)
@click.option('--shingle_size', help="what shingle/n-gram size do you want for cosine similarity calculations?",
              default=2, show_default=True)
@click.option('--max_cosine',
              help="""how much of a cosine distance will you tolerate 
              when comparing an enum name to a term lable or synonym?""",
              default=0.05, show_default=True)
@click.option('--query_field_string',
              help="""do you want to define a custom list of fields to search in? 
              The default settings work well in most cases.""",
              default='', show_default=True)
@click.option('--test_sample_size',
              help="""if greater than 0, the enum name list will be samples at this size before mapping.""",
              default=0, show_default=True)
@click.option('--trim_parentheticals/--keep_parentheticals', default=False)
def enum_annotator(modelfile, all_mappings_fn, requested_enum_name, whiteout_chars, ontology_string,
                   ols_search_base_url, ols_terms_based_url, desired_row_count, shingle_size, max_cosine,
                   overwrite_meaning, query_field_string, test_sample_size, trim_parentheticals):
    # show entire width of data frames
    pd.set_option('display.expand_frame_repr', False)

    # GLOBALS within this method
    blank_row = {'title': '', 'id': '', 'iri': '', 'is_defining_ontology': '',
                 'label': '', 'obo_id': '', 'ontology_name': '', 'ontology_prefix': '',
                 'short_form': '', 'type': ''}

    # ols_annotations_cols = ['name', 'obo_id', 'scope', 'type', 'xrefs']

    parsed_yaml = parse_yaml_file(modelfile)

    current_schema = dict_to_schema(parsed_yaml)

    requested_enum_obj = request_an_enum(current_schema, requested_enum_name)

    requested_pvs_obj = request_pvs(requested_enum_obj)

    requested_pvs_names = get_pv_names(requested_pvs_obj)
    requested_pvs_names.sort()
    logger.debug(requested_pvs_names)

    ontologies_phrased = make_ontolgy_phrase(ontology_string)
    logger.debug(ontologies_phrased)

    qf_phrased = make_qf_phrase(query_field_string)
    logger.debug(qf_phrased)

    cosine_obj = make_cosine_obj(shingle_size)
    # logger.debug(cosine_obj)

    # initialize
    enum_name_mappings = DataFrameClass()
    # logger.debug(enum_name_mappings.get())
    term_annotations = DataFrameClass()
    # logger.debug(term_annotations.get())

    reusable_session = requests.Session()
    # logger.debug(reusable_session)

    # # load with schemaview and extract SchemaDefinition?
    # # or open directly into a SchemaDefinition?  HOW?
    # # https://fedingo.com/how-to-read-yaml-file-to-dict-in-python/

    if test_sample_size > 0:
        requested_pvs_names = random.sample(requested_pvs_names, test_sample_size)
        requested_pvs_names.sort()

    for pv_name in requested_pvs_names:
        logger.info(pv_name)
        # current_pv = requested_pvs_obj[pv_name]
        # logger.debug(current_pv)
        ols_term_search(pv_name, whiteout_chars, ontologies_phrased, qf_phrased, desired_row_count,
                        blank_row, enum_name_mappings, reusable_session, ols_search_base_url, trim_parentheticals)
        # # returns true
        # # could look at growth in enum_name_mappings

        enum_name_mapping_frame = enum_name_mappings.get()
        # logger.debug(enum_name_mapping_frame)

        term_and_source = enum_name_mapping_frame.loc[enum_name_mapping_frame['raw_query'].eq(pv_name)]

        term_and_source = term_and_source[["iri", "ontology_name"]]
        term_and_source.drop_duplicates(inplace=True)
        term_and_source = term_and_source.loc[~term_and_source['iri'].eq("")]
        term_and_source.sort_values(["iri", "ontology_name"], inplace=True)
        logger.debug(term_and_source)

        term_and_source = term_and_source.to_dict(orient="records")

        for i in term_and_source:
            # returns true and saves to a dataframe class
            get_ols_term_annotations(i["iri"], i["ontology_name"], reusable_session,
                                     ols_terms_based_url, term_annotations)

        annotations_from_terms = term_annotations.get()

        # logger.warning(enum_name_mapping_frame)
        # # warning:    tidied_query     raw_query title id iri is_defining_ontology label obo_id ontology_name ontology_prefix short_form type
        # # warning: 0  metabolomics  metabolomics
        # logger.warning(annotations_from_terms)
        # # warning: Empty DataFrame
        # # warning: Columns: []
        # # warning: Index: []

        if len(enum_name_mapping_frame.index) > 0 and len(annotations_from_terms.index) > 0:
            raw_through_annotations = enum_name_mapping_frame.merge(annotations_from_terms, how='left', on="iri",
                                                                    suffixes=('_term', '_ano'))

            for_str_dist = raw_through_annotations[["tidied_query", "name"]]
            # 20211215 0912
            # A value is trying to be set on a copy of a slice from a DataFrame.
            for_str_dist["tidied_query_lc"] = for_str_dist["tidied_query"].str.lower()
            for_str_dist["name_lc"] = for_str_dist["name"].str.lower()

            # favoring simplicity over efficiency
            # ie may be string-comparing some duplicates
            # easier to merge back in
            # for_str_dist = for_str_dist.loc[
            #     ~for_str_dist["tidied_query"].eq("") and ~for_str_dist["label"].eq("") and ~for_str_dist[
            #         "tidied_query"].isnull() and ~for_str_dist["label"].isnull()]
            # for_str_dist.drop_duplicates(inplace=True)
            # for_str_dist.sort_values(["tidied_query", "label"], inplace=True)

            for_str_dist_dict = for_str_dist.to_dict(orient="records")

            # dist_list = []
            new_pair_list = []

            for pair in for_str_dist_dict:
                # used to get_profile
                name_type = type(pair["name"])
                if name_type is str:
                    the_dist = cosine_obj.distance(pair["tidied_query_lc"], pair["name_lc"])
                    pair['cosine'] = the_dist
                else:
                    pair['cosine'] = None
                new_pair_list.append(pair)

            for_str_dist = pd.DataFrame(new_pair_list)
            for_str_dist.drop(labels=["tidied_query_lc", "name_lc"], axis=1, inplace=True)

            # was debug
            logger.debug(for_str_dist)
            raw_through_dist = raw_through_annotations.merge(for_str_dist, how="left", on=["tidied_query", "name"])
            all_mappings_frame = []

            new_enum = linkml_runtime.linkml_model.EnumDefinition(name=requested_enum_name)
            # logger.info(new_enum)

            # looping inside the same loop ?!
            for i in requested_pvs_names:
                # todo unnest loop?
                logger.debug(i)
                ce = requested_pvs_obj[i]
                cr = raw_through_dist.loc[raw_through_dist["raw_query"].eq(i)]
                all_mappings_frame.append(cr)
                cr_row_count = len(cr.index)
                if cr_row_count > 0:
                    min_cosine = cr["cosine"].min()
                    with_min = cr.loc[cr["cosine"] == min_cosine]
                    with_min_row_count = len(with_min.index)
                    if with_min_row_count > 0:
                        with_min = with_min.drop(labels=['xrefs'], axis=1)
                        if 'description' in with_min.columns:
                            with_min['description'] = str(with_min['description'])
                        with_min.drop_duplicates(inplace=True)
                        deduped_row_count = len(with_min.index)
                        # # I'm surprised that there aren't any 2+ equally good mappings here
                        # will have to deal with that at some point
                        # may still need to do some row filtering/prioritizing by source or annotation type
                        # prefer label over synonym
                        # Prefer ontologies in the order they appear in XXX parameter?
                        if deduped_row_count > 1:
                            pass
                        first_row_as_dict = (with_min.to_dict(orient="records"))[0]
                        # print(first_row_as_dict)
                        ce.annotations["match_val"] = first_row_as_dict['name']
                        ce.annotations["match_type"] = first_row_as_dict['scope']
                        ce.annotations["match_id"] = first_row_as_dict['obo_id_term']
                        ce.annotations["match_pref_lab"] = first_row_as_dict['pref_lab']
                        ce.annotations["cosine"] = first_row_as_dict['cosine']
                        if overwrite_meaning:
                            if first_row_as_dict['cosine'] <= max_cosine:
                                ce.meaning = first_row_as_dict['obo_id_term']
                                ce.title = first_row_as_dict['label']
                            else:
                                ce.meaning = None
                                ce.title = None
                        new_enum.permissible_values[i] = ce

    all_mappings_frame = pd.concat(all_mappings_frame)

    all_mappings_frame.to_csv(all_mappings_fn, sep="\t", index=False)

    if 'description' in all_mappings_frame.columns:
        all_mappings_frame['description'] = str(all_mappings_frame['description'])
    if 'xrefs' in all_mappings_frame.columns:
        all_mappings_frame['xrefs'] = str(all_mappings_frame['xrefs'])

    all_mappings_frame.drop_duplicates(inplace=True)

    all_mappings_frame.to_csv(all_mappings_fn, sep="\t", index=False)

    gs_tq_vc = all_mappings_frame["raw_query"].value_counts()
    logger.info("Number of hits for each term, regardless of quality. Doesn't show terms with 0 hits.")
    logger.info(gs_tq_vc)

    current_schema.enums[requested_enum_name] = new_enum

    string_dumped_schema = yaml_dumper.dumps(current_schema)

    # todo send to STDOUT or output file?
    print(string_dumped_schema)
    # # with open(xxx, 'w') as file:
    # #     documents = yaml.safe_dump(string_dumped_schema, file)


if __name__ == '__main__':
    enum_annotator()



from schema_automator.importers.jsonschema_import_engine import JsonSchemaImportEngine
from schema_automator.importers.owl_import_engine import OwlImportEngine
from schema_automator.importers.dosdp_import_engine import DOSDPImportEngine




from dataclasses import dataclass


@dataclass
class GeneralSchemaEnhancer:
    """
    Main functions have moved to core linkml, see https://github.com/linkml/linkml/pull/854

    This is currently a stub for future enhancements
    """
    pass









import logging
import click
import click_log
import pandas as pd
from linkml_runtime.utils.schemaview import SchemaView

logger = logging.getLogger(__name__)
click_log.basic_config(logger)


@click.command()
@click_log.simple_verbosity_option(logger)
@click.option('--modelfile', help="path to LinkML input", type=click.Path(exists=True), show_default=True,
              required=True)
@click.option('--output_file', help="path to TSV output", type=click.Path(), required=True)
def enums_pvs_tsv(modelfile, output_file):
    """
    Writes a TSV report of the enums and permissible values in modelfile (a LinkML YAML file) to output_file (a TSV file).
    """
    row_list = []
    view = SchemaView(modelfile)
    enums = view.all_enums()
    for ek, ev in enums.items():
        pvs = ev.permissible_values
        for pk, pv in pvs.items():
            logger.debug(f"{ek}: {pk}")
            row_list.append({"enum": ek, "pv": pk})
    row_frame = pd.DataFrame(row_list)
    row_frame.to_csv(output_file, sep="\t", index=False)


if __name__ == '__main__':
    enums_pvs_tsv()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['schema_automator',
 'schema_automator.annotators',
 'schema_automator.dosdp',
 'schema_automator.enhancer',
 'schema_automator.generalizers',
 'schema_automator.importers',
 'schema_automator.jsonschema',
 'schema_automator.utils']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.3.1,<6.0.0',
 'bioregistry>=0.4.59,<0.5.0',
 'click-log>=0.4.0,<0.5.0',
 'click>=8.1.2,<9.0.0',
 'funowl>=0.1.11,<0.2.0',
 'jsonpatch>=1.32,<2.0',
 'linkml-runtime>=1.2.17,<2.0.0',
 'linkml>=1.2.11,<2.0.0',
 'mkdocs>=1.2.3,<2.0.0',
 'oaklib>=0.1.17,<0.2.0',
 'pandas>=1.3.5,<2.0.0',
 'psycopg2-binary>=2.9.2,<3.0.0',
 'pypandoc>=1.7.5,<2.0.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'quantulum3>=0.7.9,<0.8.0',
 'rdflib>=6.0.2,<7.0.0',
 'requests>=2.26.0,<3.0.0',
 'strsimpy>=0.2.1,<0.3.0',
 'twine>=4.0.0,<5.0.0']

entry_points = \
{'console_scripts': ['curated_to_enums = '
                     'schema_automator.annotators.curated_to_enums:curated_to_enums',
                     'dosdp2linkml = '
                     'schema_automator.importers.owl_import_engine:dosdp2model',
                     'enum_annotator = '
                     'schema_automator.annotators.enum_annotator:enum_annotator',
                     'enums_to_curateable = '
                     'schema_automator.annotators.enums_to_curateable:enums_to_curateable',
                     'extract-schema = '
                     'schema_automator.utils.schema_extractor:cli',
                     'jsondata2linkml = '
                     'schema_automator.importers.json_instance_import_engine:json2model',
                     'jsonschema2linkml = '
                     'schema_automator.importers.jsonschema_import_engine:jsonschema2model',
                     'owl2linkml = '
                     'schema_automator.importers.owl_import_engine:owl2model',
                     'rdf2linkml = '
                     'schema_automator.importers.rdf_instance_import_engine:rdf2model',
                     'schemauto = schema_automator.cli:main',
                     'tsv2linkml = '
                     'schema_automator.importers.csv_import_engine:tsv2model',
                     'tsvs2linkml = '
                     'schema_automator.importers.csv_import_engine:tsvs2model']}

setup_kwargs = {
    'name': 'schema-automator',
    'version': '0.2.4',
    'description': 'Infer models, enrich with meaning for terms including enum permissible values',
    'long_description': '# LinkML Schema Automator\n\nThis is a toolkit that assists with the generation and enhancement of [LinkML](https://linkml.io/linkml) schemas\n\n## Install\n\n```bash\npip install schema-automator\n```\n\n## Command Line\n\nSee [CLI docs](https://linkml.io/schema-automator/cli)\n\nGeneralizing:\n\n```bash\nschemauto generalize-tsv my-data.tsv > my-schema.yaml\n```\n\nImporting:\n\n```bash\nschemauto import-json-schema their.schema.json > my-schema.yaml\n```\n\nAnnotating:\n\n```bash\nschemauto annotate-schema my-schema.yaml\n```\n\n## Full Documentation\n\n[linkml.io/schema-automator](https://linkml.io/schema-automator/)\n',
    'author': 'Chris Mungall',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import urllib.request, urllib.error, urllib.parse

import click
import requests
import logging
import json
import yaml
import os
from dataclasses import dataclass
from pprint import pprint
from typing import Any, List, Dict, Union, Iterator

from linkml_runtime.linkml_model import SchemaDefinition
from linkml_runtime.utils.metamodelcore import Curie
from linkml_runtime.utils.schemaview import SchemaView, re
from oaklib import BasicOntologyInterface
from oaklib.datamodels.search import SearchConfiguration
from oaklib.datamodels.text_annotator import TextAnnotation
from oaklib.interfaces import SearchInterface
from oaklib.interfaces.text_annotator_interface import TextAnnotatorInterface

from schema_automator.utils.schemautils import minify_schema

REST_URL = "http://data.bioontology.org"
camel_case_pattern = re.compile(r'(?<!^)(?=[A-Z])')

def uncamel(n: str):
    return camel_case_pattern.sub(' ', n).lower().replace('_', ' ')

@dataclass
class SchemaAnnotator:
    ontology_implementation: BasicOntologyInterface

    def annotate_text(self, text: str) -> Iterator[TextAnnotation]:
        # this is a wrapper over OAK annotation and search;
        # it (1) expands CamelCase (2) abstracts over annotation vs search
        # TODO: fold this functionality back into OAK
        oi = self.ontology_implementation
        text_exp = uncamel(text)
        if isinstance(oi, TextAnnotatorInterface):
            # TextAnnotation is available; use this by default
            for r in oi.annotate_text(text_exp):
                yield r
            if text_exp != text.lower():
                for r in oi.annotate_text(text_exp):
                    yield r
        elif isinstance(oi, SearchInterface):
            # use search as an alternative
            cfg = SearchConfiguration(is_complete=True)
            for r in oi.basic_search(text, config=cfg):
                yield TextAnnotation(object_id=r, matches_whole_text=True)
            if text_exp != text.lower():
                for r in oi.basic_search(text_exp, config=cfg):
                    yield TextAnnotation(object_id=r, matches_whole_text=True)
        else:
            raise NotImplementedError

    def annotate_schema(self, schema: Union[SchemaDefinition, str], curie_only=True) -> SchemaDefinition:
        """
        Annotate all elements of a schema, adding mappings
        """
        sv = SchemaView(schema)
        oi = self.ontology_implementation
        for elt_name, elt in sv.all_elements().items():
            for n in [elt.name] + elt.aliases:
                for r in self.annotate_text(n):
                    logging.debug(f'MATCH: {r}')
                    if r.matches_whole_text:
                        xref = r.object_id
                        if curie_only and not Curie.is_curie(xref):
                            continue
                        logging.info(f'Mapping from {elt_name} "{n}" to {xref}')
                        if xref not in elt.exact_mappings:
                            elt.exact_mappings.append(xref)
        for e in sv.all_enums().values():
            for pv in e.permissible_values.values():
                for r in self.annotate_text(pv.text):
                    logging.debug(f'MATCH: {r}')
                    if r.matches_whole_text:
                        xref = r.object_id
                        if curie_only and not Curie.is_curie(xref):
                            continue
                        logging.info(f'Mapping from {elt_name} "{n}" to {xref}')
                        if pv.meaning is None:
                            logging.info(f'Arbitrarily choosing first match: {xref}')
                            pv.meaning = xref
                        else:
                            if xref not in pv.exact_mappings:
                                pv.exact_mappings.append(xref)

        return sv.schema


@click.command()
@click.argument('schema')
@click.option('--input', '-i', help="OAK input ontology selector")
@click.option('--output', '-o', help="Path to saved yaml schema")
def annotate_schema(schema: str, input: str, output: str, **args):
    """
    Annotate all elements of a schema
    """
    logging.basicConfig(level=logging.INFO)
    annr = SchemaAnnotator()
    schema = annr.annotate_schema(schema)
    sd = minify_schema(schema)
    if output:
        with open(output, 'w') as stream:
            yaml.safe_dump(sd, stream, sort_keys=False)
    else:
        print(yaml.safe_dump(sd, sort_keys=False))

if __name__ == '__main__':
    annotate_schema()


import click
import logging
import yaml
from typing import Dict, List, Optional
from collections import defaultdict
import os
import re
import csv
import requests
import pandas as pd
import time

from dateutil.parser import parse
from linkml_runtime import SchemaView
from linkml_runtime.linkml_model import SchemaDefinition, ClassDefinition, TypeDefinition, SlotDefinition
from quantulum3 import parser as q_parser
from dataclasses import dataclass, field

from schema_automator.generalizers.generalizer import Generalizer
from schema_automator.importers.import_engine import ImportEngine
from schema_automator.utils.schemautils import merge_schemas, write_schema

ID_SUFFIX = '_id'

ROBOT_NAME_MAP = {
    'alternative term': "IAO:0000118",
    'definition': "IAO:0000115",
    'definition_source': "IAO:0000119",
}


@dataclass
class ForeignKey:
    """
    Represents a field in one table that points to an identifier field in another
    """
    source_table: str
    source_column: str
    target_table: str
    target_column: str
    num_distinct_values: int
    range: str = None

    def score(self):
        s = 0
        if self.source_table == self.target_table:
            s -= 1
        if self.source_column.lower().endswith(ID_SUFFIX):
            s += 1
        if self.target_column.lower().endswith(ID_SUFFIX) or self.target_column.lower() == 'id':
            s += 1
        if self.source_column.lower().startswith(self.target_table.lower()):
            s += 2
        if self.range != 'string' and self.range != 'integer':
            s -= 1
        return s


@dataclass
class CsvDataGeneralizer(Generalizer):
    """
    A Generalizer that generalizes from example CSV/TSV data
    """

    column_separator: str = "\t"
    schema_name: str = 'example'
    robot: bool = False
    enum_columns: List[str] = field(default_factory=lambda: [])
    enum_mask_columns: List[str] = field(default_factory=lambda: [])
    enum_threshold: float = 0.1
    enum_strlen_threshold: int = 30
    max_enum_size: int = 50
    downcase_header: bool = False
    infer_foreign_keys: bool = False
    max_pk_len: int = 60   # URIs can be long..
    min_distinct_fk_val: int = 8

    def infer_linkages(self, files: List[str], **kwargs) -> List[ForeignKey]:
        """
        Heuristic procedure for determining which tables are linked to others via implicit foreign keys

        If all values of one column FT.FC are present in column PT.PC, then FT.FC is a potential foreign key
        and PC is a potential primary key of PT.

        This procedure can generate false positives, so additional heuristics are applied. Each potential
        foreign key relationship gets an ad-hoc score:
         - links across tables score more highly than within
         - suffixes such as _id are more likely on PK and FK tables
         - the foreign key column table is likely to start with the base column name
        In addition, if there are competing primary keys for a table, the top scoring one is selected
        """
        fks: List[ForeignKey] = []
        MAX_PK_LEN = self.max_pk_len
        dfs: Dict[str, pd.DataFrame] = {}

        for file in files:
            c = os.path.splitext(os.path.basename(file))[0]
            if self.downcase_header:
                c = c.lower()
            logging.info(f'READING {file} ')
            df = pd.read_csv(file, sep=self.column_separator, skipinitialspace=True).fillna("")
            if self.downcase_header:
                df = df.rename(columns=str.lower)
            exclude = []
            for col in df.columns:
                vals = set(df[col].tolist())
                if len(vals) < self.min_distinct_fk_val:
                    logging.info(f'EXCLUDING {col} (too few, len = {len(vals)})')
                    exclude.append(col)
                    continue
                max_str_len = max([len(str(x)) for x in vals])
                if max_str_len > MAX_PK_LEN:
                    logging.info(f'EXCLUDING {col} (len {max_str_len} > {MAX_PK_LEN}) sample: {list(vals)[0:5]}')
                    #for v in vals:
                    #    if len(str(v)) == max_str_len:
                    #        print(f'  WITNESS: {v}')
                    exclude.append(col)
                    continue
                if any(' ' in str(x) for x in vals ):
                    logging.info(f'EXCLUDING {col} (has spaces)')
                    exclude.append(col)
                    continue
            for col in exclude:
                del df[col]
                logging.info(f'Excluding: {col}')
            dfs[c] = df
        for t_primary, df_primary in dfs.items():
            for candidate_pk in df_primary.columns:
                candidate_pk_vals = set(df_primary[candidate_pk].tolist())
                candidate_pk_range = infer_range({}, candidate_pk_vals, {})
                logging.info(f'Candidate PK {t_primary}.{candidate_pk} ')
                for t_foreign, df_foreign in dfs.items():
                    logging.info(f' Candidate FK table {t_foreign} ')
                    for candidate_fk in df_foreign.columns:
                        logging.info(f'  Candidate FK col {candidate_fk} ')
                        if t_primary == t_foreign and candidate_pk == candidate_fk:
                            logging.info(f'   SKIP (identical) {candidate_fk} ')
                            continue
                        candidate_fk_vals = set(df_foreign[candidate_fk].tolist())
                        logging.info(f'    Candidate FK {t_foreign}.{candidate_fk}')
                        is_fk = True
                        for v in candidate_fk_vals:
                            if v is None or v == '':
                                continue
                            if v not in candidate_pk_vals:
                                logging.info(f'    {v} not in candidates')
                                is_fk = False
                            if not is_fk:
                                break
                        if is_fk:
                            logging.info(f'    all {len(candidate_fk_vals)} fk vals in {len(candidate_pk_vals)} pks')
                            fks.append(ForeignKey(source_table=t_foreign,
                                                  source_column=candidate_fk,
                                                  target_table=t_primary,
                                                  target_column=candidate_pk,
                                                  num_distinct_values=len(candidate_fk_vals),
                                                  range=candidate_pk_range))

        pk_tables = set([fk.target_table for fk in fks])
        for pk_table in pk_tables:
            s = defaultdict(float)
            max_s = -1000
            for fk in fks:
                if fk.target_table == pk_table:
                    s[fk.target_column] += fk.score()
                    if s[fk.target_column] > max_s:
                        max_s = s[fk.target_column]
            pk_col, _ = sorted(s.items(), key=lambda item: -item[1])[0]
            logging.info(f'SELECTED pk col {pk_col} for {pk_table}; scores = {s}')
            fks = [fk for fk in fks if not (fk.target_table == pk_table and fk.target_column != pk_col)]
        fks = [fk for fk in fks if fk.score() > 0]
        logging.info(f'FILTERED: {fks}')
        return fks

    def inject_foreign_keys(self, sv: SchemaView, fks: List[ForeignKey]) -> None:
        schema = sv.schema
        for fk in fks:
            # TODO: deal with cases where the same slot is used in different classes
            src_cls = schema.classes[fk.source_table]
            src_slot = schema.slots[fk.source_column]
            src_cls.slot_usage[fk.source_column] = \
                SlotDefinition(name=fk.source_column,
                               range=fk.target_table)
            #src_slot['range'] = fk.target_table
            tgt_cls = schema.classes[fk.target_table]
            tgt_slot = schema.slots[fk.target_column]
            tgt_cls.slot_usage[fk.target_column] = \
                SlotDefinition(name=fk.target_column,
                               identifier=True)
            #tgt_slot['identifier'] = True

    def convert_multiple(self, files: List[str], **kwargs) -> SchemaDefinition:
        """
        Converts multiple TSVs to a schema

        :param files:
        :param kwargs:
        :return:
        """
        if self.infer_foreign_keys:
            fks = self.infer_linkages(files)
        else:
            fks = ()
        schemas = []
        for file in files:
            c = os.path.splitext(os.path.basename(file))[0]
            if self.downcase_header:
                c = c.lower()
            s = self.convert(file, class_name=c, **kwargs)
            if s is not None:
                schemas.append(s)
            logging.info(f'Classes={list(s.classes.keys())}')
        sv = SchemaView(schemas[0])
        for s in schemas[1:]:
            sv.merge_schema(s)
            logging.info(f'Classes, post merge={list(sv.all_classes().keys())}')
        #s = merge_schemas(yamlobjs)
        self.inject_foreign_keys(sv, fks)
        return sv.schema

    def convert(self, file: str, **kwargs) -> SchemaDefinition:
        """
        Converts a single TSV file to a single-class schema
        
        :param file:
        :param kwargs:
        :return:
        """
        with open(file, newline='') as tsv_file:
            header = [h.strip() for h in tsv_file.readline().split('\t')]
            rr = csv.DictReader(tsv_file, fieldnames=header, delimiter=self.column_separator, skipinitialspace=False)
            return self.convert_dicts([r for r in rr], **kwargs)

    def read_slot_tsv(self, file: str, **kwargs) -> Dict:
        with open(file, newline='') as tsv_file:
            rows_list = csv.reader(tsv_file, delimiter=self.column_separator)
            return self.convert_to_edge_slots([r for r in rows_list], **kwargs)

    def convert_to_edge_slots(self,
                         all_tsv_rows: List,
                         name: str = 'example',
                         **kwargs) -> Optional[Dict]:

        """
        assume that TSV has 3 relevant columns:
            1. slot name to add
            2. slot definition to add
            3. examples of values for the slot

        also assume that these are all edge_properties at the moment.
        TODO: add parameter to allow edge or node property disambiguation.
        """

        slots = {}
        slot_values = {}
        types = {}

        for item in all_tsv_rows:
            slot_name = item[0]
            slot_definition = item[1]
            slot_example_type = item[2]
            if isinstance(slot_example_type, list):
                vs = slot_example_type
            elif isinstance(slot_example_type, str):
                vs = slot_example_type.split('|')
            else:
                vs = [slot_example_type]
            if slot_name not in slots:
                slots[slot_name] = {'is_a': 'association slot', 'description': slot_definition, 'range': None}
                slot_values[slot_name] = set()
            if slot_example_type is not None and slot_example_type != "" and not str(slot_example_type).startswith('$ref:'):
                slots[slot_name]['examples'] = [{'value': slot_example_type}]
                slot_values[slot_name].update(vs)
            if len(vs) > 1:
                slots[slot_name]['multivalued'] = True

        new_slots = {}
        # slots is a dict{dict}
        for sn, s in slots.items():
            vals = slot_values[sn]
            s['range'] = infer_range(s, vals, types)
        for sn, s in new_slots.items():
            if sn not in slots:
                slots[sn] = s
        schema = {
            'slots': slots
        }
        return schema

    def convert_dicts(self,
                      rr: List[Dict],
                      schema_name: str = 'example',
                      class_name: str = 'example',
                      **kwargs) -> SchemaDefinition:
        slots = {}
        slot_values = {}
        n = 0
        enums = {}
        robot_defs = {}
        slot_usage = {}
        types = {}
        enum_columns = self.enum_columns
        enum_mask_columns = self.enum_mask_columns
        if len(rr) == 0:
            return None
        for row in rr:
            if self.downcase_header:
                row = {k.lower(): v for k, v in row.items()}
            n += 1
            if n == 1 and self.robot:
                for k, v in row.items():
                    robot_defs[k] = v
                continue
            for k, v in row.items():
                if k is None or k == '':
                    continue
                if v is None:
                    v = ""
                if isinstance(v, str):
                    v = v.strip()
                if isinstance(v, list):
                    vs = v
                elif isinstance(v, str):
                    vs = v.split('|')
                else:
                    vs = [v]
                if k not in slots:
                    slots[k] = {'range': None}
                    slot_values[k] = set()
                if v is not None and v != "" and not str(v).startswith('$ref:'):
                    slots[k]['examples'] = [{'value': v}]
                    slot_values[k].update(vs)
                if len(vs) > 1:
                    slots[k]['multivalued'] = True
        types = {}
        new_slots = {}
        for sn, s in slots.items():
            vals = slot_values[sn]
            s['range'] = infer_range(s, vals, types)
            if (s['range'] == 'string' or sn in enum_columns) and sn not in enum_mask_columns:
                n_distinct = len(vals)
                longest = max([len(str(v)) for v in vals]) if n_distinct > 0 else 0
                if sn in enum_columns or \
                        ((n_distinct / n) < self.enum_threshold and 0 < n_distinct <= self.max_enum_size
                         and longest < self.enum_strlen_threshold):
                    enum_name = sn.replace(' ', '_').replace('(s)', '')
                    enum_name = f'{enum_name}_enum'
                    s['range'] = enum_name
                    enums[enum_name] = {
                        'permissible_values': {v:{'description': v} for v in vals}
                    }
            # ROBOT template hints. See http://robot.obolibrary.org/template
            if sn in robot_defs:
                rd = robot_defs[sn]
                if 'SPLIT' in rd:
                    rd = re.sub(' SPLIT.*', '', rd)
                if rd.startswith("EC"):
                    rd = rd.replace('EC ', '')
                    rel = capture_robot_some(rd)
                    ss = rd.replace('%', '{' + sn + '}')
                    slot_usage['equivalence axiom'] = {'string_serialization': ss}
                    if rel is not None:
                        s['is_a'] = rel
                        new_slots[rel] = {}
                elif rd.startswith("SC"):
                    rd = rd.replace('SC ', '')
                    rel = capture_robot_some(rd)
                    ss = rd.replace('%', '{' + sn + '}')
                    slot_usage['subclass axiom'] = {'string_serialization': ss}
                    if rel is not None:
                        s['is_a'] = rel
                        new_slots[rel] = {}
                        s['comments'] = ['OWL>> SomeValuesFrom']
                    else:
                        s['comments'] = ['OWL>> SubClassOf']
                elif rd.startswith("C"):
                    # TODO: semantics are dependent on CLASS_TYPE column
                    # https://robot.obolibrary.org/template
                    rd = rd.replace('C ', '')
                    if rd == '%':
                        s['broad_mappings'] = ['rdfs:subClassOf']
                    rel = capture_robot_some(rd)
                    if rel is not None:
                        s['is_a'] = rel
                        new_slots[rel] = {}
                elif rd == 'ID':
                    s['identifier'] = True
                elif rd.startswith("I"):
                    rd = rd.replace('I ', '')
                    # TODO
                elif rd == 'TYPE':
                    s['slot_uri'] = 'rdf:type'
                elif rd == 'LABEL':
                    s['slot_uri'] = 'rdfs:label'
                elif rd.startswith("A "):
                    s['slot_uri'] = rd.replace('A ', '')
                elif rd.startswith("AT "):
                    s['slot_uri'] = re.sub('^^.*', '', rd.replace('AT ', ''))
                elif rd.startswith(">A "):
                    logging.warning('Axiom annotations not supported')
                slot_uri = s.get('slot_uri', None)
                if slot_uri is not None:
                    if ' ' in slot_uri or ':' not in slot_uri:
                        if slot_uri in ROBOT_NAME_MAP:
                            s['slot_uri'] = ROBOT_NAME_MAP[slot_uri]
                        else:
                            del s['slot_uri']
                            logging.warning(f'ROBOT "A" annotations not supported yet')
        class_slots = list(slots.keys())
        for sn, s in new_slots.items():
            if sn not in slots:
                slots[sn] = s
        schema = SchemaDefinition(
            id=f'https://w3id.org/{schema_name}',
            name=schema_name,
            description=schema_name,
            imports=['linkml:types'],
            default_prefix=schema_name,
            types=types,
            classes=[
                ClassDefinition(class_name,
                                slots=class_slots,
                                slot_usage=slot_usage)
            ],
            slots=slots,
            enums=enums
        )
        self.add_default_prefixes(schema)
        self.add_prefix(schema, schema_name, f'https://w3id.org/{schema_name}')
        if robot_defs:
            self.add_prefix(schema, 'IAO', 'http://purl.obolibrary.org/obo/IAO_')
        add_missing_to_schema(schema)
        return schema


def capture_robot_some(s: str) -> str:
    """
    parses an OWL some values from from a robot template

    :param s:
    :return:
    """
    results = re.findall('(\\S+) some %',s)
    if len(results) == 0:
        return None
    else:
        r = results[0]
        if ':' in r:
            # only use named properties
            return None
        else:
            return r


def isfloat(value):
    try:
        float(value)
        return True
    except ValueError:
        return False


def is_measurement(value):
    ms = q_parser.parse(value)
    for m in ms:
        if m.unit.name != 'dimensionless':
            return True


def is_all_measurement(values):
    """
    heuristic to guess if all values are measurements

    uses quantulum to parse

    A significant proportion must be dimensional, to avoid
    accidentally classifying a list of floats as measurements
    """
    n_dimensional = 0
    n = 0
    for value in values:
        if not isinstance(value, str):
            return False
        try:
            ms = q_parser.parse(value)
        except:
            return False
        if len(ms) == 0:
            return False
        n += 1
        if all(m.unit.name != 'dimensionless' for m in ms):
            n_dimensional += 1
    # TODO: make this configurable
    if n_dimensional > n/2:
        return True
    else:
        return False


def infer_range(slot: dict, vals: set, types: dict) -> str:
    nn_vals = [v for v in vals if v is not None and v != ""]
    if len(nn_vals) == 0:
        return 'string'
    if all(str(v).startswith('$ref:') for v in nn_vals):
        return nn_vals[0].replace('$ref:', '')
    if all(isinstance(v, int) for v in nn_vals):
        return 'integer'
    if all(isinstance(v, float) for v in nn_vals):
        return 'float'
    if all(str(v).isdigit() for v in nn_vals):
        return 'integer'
    if all(is_date(v) for v in nn_vals):
        return 'datetime'
    if all(isfloat(v) for v in nn_vals):
        return 'float'
    if is_all_measurement(nn_vals):
        return 'measurement'
    v0 = nn_vals[0]
    db = get_db(v0)
    if db is not None:
        if all(get_db(v) == db for v in nn_vals):
            t = f'{db} identifier'
            types[t] = {'typeof': 'string'}
            return t
        if all(get_db(v) is not None for v in nn_vals):
            t = 'identifier'
            types[t] = {'typeof': 'string'}
            return t
    return 'string'


def get_db(db_id: str) -> str:
    parts = db_id.split(':')
    if len(parts) > 1:
        return parts[0]
    else:
        return None


def is_date(string, fuzzy=False):
    """
    Return whether the string can be interpreted as a date.

    :param string: str, string to check for date
    :param fuzzy: bool, ignore unknown tokens in string if True
    """
    try:
        parse(string, fuzzy=fuzzy)
        return True
    except Exception:
        # https://stackoverflow.com/questions/4990718/how-can-i-write-a-try-except-block-that-catches-all-exceptions
        # we don't know all the different parse exceptions, we assume any error means this is a date
        return False


@dataclass
class Hit:
    term_id: str
    name: str
    score: float


def get_pv_element(v: str, zooma_confidence: str, cache: dict = {}) -> Hit:
    """
    uses ZOOMA to guess a meaning of an enum permissible value

    :param v:
    :param zooma_confidence:
    :param cache:
    :return:
    """
    if v in cache:
        return cache[v][0]
    if zooma_confidence is None:
        return None

    def confidence_to_int(c: str) -> int:
        if c == 'HIGH':
            return 5
        elif c == 'GOOD':
            return 4
        elif c == 'MEDIUM':
            return 2
        elif c == 'LOW':
            return 1
        else:
            raise Exception(f'Unknown: {c}')
    confidence_threshold = confidence_to_int(zooma_confidence)

    ontscores = {
        'NCBITaxon': 1.0,
        'OMIT': -1.0,

    }

    # zooma doesn't seem to do much pre-processing, so we convert
    label = v
    if 'SARS-CoV' not in label:
        label = re.sub("([a-z])([A-Z])", "<1> <2>", label)  # expand CamelCase
    label = label.replace('.', ' ').replace('_', ' ')
    params = {'propertyValue': label}
    time.sleep(1)  # don't overload service
    logging.info(f'Q: {params}')
    r = requests.get('http://www.ebi.ac.uk/spot/zooma/v2/api/services/annotate',params=params)
    hits = []  # List[hit]
    for hit in r.json():
        confidence = float(confidence_to_int(hit['confidence']))
        id = hit['semanticTags'][0]
        if confidence >= confidence_threshold:
            hit = Hit(term_id=id,
                      name=hit['annotatedProperty']['propertyValue'],
                      score=confidence)
            hits.append(hit)
        else:
            logging.warning(f'Skipping {id} {confidence}')
    hits = sorted(hits, key=lambda h: h.score, reverse=True)
    logging.info(f'Hits for {label} = {hits}')
    if len(hits) > 0:
        cache[label] = hits
        return hits[0]
    else:
        return None


def convert_range(k: str, dt: str) -> str:
    t = 'string'
    if dt == 'float64':
        t = 'float'
    return t


def infer_enum_meanings(schema: dict,
                        zooma_confidence: str = 'MEDIUM',
                        cache={}) -> None:
    for _,e in schema['enums'].items():
        pvs = e['permissible_values']
        for k, pv in pvs.items():
            if pv is None:
                pv = {}
                pvs[k] = pv
            if 'meaning' not in pv or pv['meaning'] is not None:
                hit = get_pv_element(k, zooma_confidence=zooma_confidence, cache=cache)
                if hit is not None:
                    pv['meaning'] = hit.term_id
                    if 'description' not in pv:
                        pv['description'] = hit.name


def add_missing_to_schema_dict(schema: dict):
    for slot in schema['slots'].values():
        if slot.get('range', None) == 'measurement':
            types = schema['types']
            if 'measurement' not in types:
                types['measurement'] = \
                    {'typeof': 'string',
                     'description': 'Holds a measurement serialized as a string'}


def add_missing_to_schema(schema: SchemaDefinition):
    for slot in schema.slots.values():
        if slot.range == 'measurement':
            types = schema.types
            if 'measurement' not in types:
                types['measurement'] = \
                    TypeDefinition('measurement',
                                   typeof='string',
                                   description='Holds a measurement serialized as a string')


@click.group()
def main():
    pass


@main.command()
@click.argument('tsvfile')  # input TSV (must have column headers
@click.option('--output', '-o', help='Output file')
@click.option('--class_name', '-c', default='example', help='Core class name in schema')
@click.option('--schema_name', '-n', default='example', help='Schema name')
@click.option('--separator', '-s', default='\t', help='separator')
@click.option('--downcase-header/--no-downcase-header', default=False, help='if true make headers lowercase')
@click.option('--enum-columns', '-E', multiple=True, help='column that is forced to be an enum')
@click.option('--robot/--no-robot', default=False, help='set if the TSV is a ROBOT template')
def tsv2model(tsvfile, output, separator, class_name, schema_name, **kwargs):
    """ Infer a model from a TSV """
    ie = CsvDataGeneralizer(**kwargs)
    schema = ie.convert(tsvfile, class_name=class_name, schema_name=schema_name)
    write_schema(schema, output)


@main.command()
@click.argument('tsvfiles', nargs=-1)  # input TSV (must have column headers
@click.option('--output', '-o', help='Output file')
@click.option('--schema_name', '-n', default='example', help='Schema name')
@click.option('--file_separator', '-s', default='\t', help='separator')
@click.option('--downcase-header/--no-downcase-header', default=False, help='if true make headers lowercase')
@click.option('--infer-foreign-keys/--no-infer-foreign-keys', default=False, help='infer ranges/foreign keys')
@click.option('--enum-columns', '-E', multiple=True, help='column(s) that is forced to be an enum')
@click.option('--enum-mask-columns', multiple=True, help='column(s) that are excluded from being enums')
@click.option('--max-enum-size', default=50, help='do not create an enum if more than max distinct members')
@click.option('--enum-threshold', default=0.1, help='if the number of distinct values / rows is less than this, do not make an enum')
@click.option('--robot/--no-robot', default=False, help='set if the TSV is a ROBOT template')
def tsvs2model(tsvfiles, output, schema_name, **kwargs):
    """ Infer a model from multiple TSVs """
    ie = CsvDataGeneralizer(**kwargs)
    schema = ie.convert_multiple(tsvfiles, schema_name=schema_name)
    write_schema(schema, output)


@main.command()
@click.argument('yamlfile')
@click.option('--zooma-confidence', '-Z', help='zooma confidence')
@click.option('--results', '-r', help='mapping results file')
def enrich(yamlfile, results, **args):
    """ Infer a model from a TSV """
    yamlobj = yaml.load(open(yamlfile))
    cache = {}
    infer_enum_meanings(yamlobj, cache=cache)
    if results is not None:
        with open(results, "w") as io:
            io.write(yaml.dump(cache))
    print(yaml.dump(yamlobj, default_flow_style=False, sort_keys=False))


if __name__ == '__main__':
    main()



import json

import click
import yaml
import logging
from copy import copy
from typing import Any, Tuple, Dict, Union, List, Optional

from linkml_runtime.linkml_model import SchemaDefinition, Element, ClassDefinition, \
    SlotDefinition, EnumDefinition, \
    ClassDefinitionName, \
    SlotDefinitionName, Prefix
from linkml_runtime.utils.formatutils import underscore

from schema_automator.importers.import_engine import ImportEngine
from schema_automator.utils.schemautils import minify_schema, write_schema

# TODO: move to core. https://github.com/linkml/linkml/issues/104
RESERVED = ['in', 'not', 'def']

class JsonSchemaImportEngine(ImportEngine):
    """
    A :ref:`ImportEngine` that imports a JSON-Schema representation to a LinkML Schema
    """

    def convert(self, input: str, name=None, format = 'json', **kwargs):
        """
        Converts a JSON-Schema json file into a LinkML schema

        :param input:
        :param name:
        :param format:
        :param kwargs:
        :return:
        """
        if format == 'json':
            with open(input) as stream:
                obj = json.load(stream)
        elif format == 'yaml':
            with open(input) as stream:
                obj = yaml.safe_load(stream)
        else:
            raise Exception(f'Bad format: {format}')
        return self.loads(obj, name, **kwargs)

    def load(self, input: str, name=None, format = 'json', **kwargs):
        return self.convert(input, name=name, format=format, **kwargs)

    def loads(self, obj: Any, name=None, **kwargs) -> SchemaDefinition:
        return self.translate_schema(obj, name, **kwargs)

    def get_id(self, obj) -> str:
        if 'id' in obj:
            id = obj['id']
        elif '$ref' in obj:
            id = obj['$ref']
        else:
            raise Exception(f'No id {obj}')
        return self.split_name(id)[0]

    def split_name(self, name) -> Tuple[str, str]:
        name = name.replace('#/definitions/', '')
        parts = name.split('.')
        name = parts.pop()
        pkg = '.'.join(parts)
        if pkg == '':
            pkg = None
        return name, pkg


    def translate_schema(self, obj: Dict, id_val=None, name=None, root_class_name=None) -> SchemaDefinition:
        if id_val is None and '$id' in obj:
            id_val = obj['$id']
        if id_val is None and '$schema' in obj:
            id_val = obj['$schema']
        if name is None and 'title' in obj:
            name = obj['title'].replace(' ','-')
        jsonschema_version = obj.get('$schema', None)
        if id_val is None and name is None:
            raise Exception(f'Must pass name OR id, or these must be present in the jsonschema')
        if name is None:
            name = id_val
        if id_val is None or not id_val.startswith('http'):
            id_val = f'https://example.org/{name}'
        name = underscore(name)
        self.schema = SchemaDefinition(id=id_val, name=name)
        if '$defs' in obj:
            self.translate_definitions(obj.get('$defs', {}))
        else:
            self.translate_definitions(obj.get('definitions', {}))
        if root_class_name is None:
            root_class_name = obj.get('title', None)
        if 'properties' in obj:
            root_class = ClassDefinition(root_class_name)
            self.translate_properties(obj, root_class)
            self.schema.classes[root_class_name] = root_class
        self.schema.default_prefix = name
        self.schema.prefixes[name] = Prefix(name, id_val)
        self.schema.prefixes['linkml'] = Prefix('linkml', 'https://w3id.org/linkml/')
        self.schema.imports.append('linkml:types')
        return self.schema

    def translate_definitions(self, obj: Dict):
        for k, v in obj.items():
            self.translate_object(v, k)

    def translate_array(self, obj: Dict, name: str) -> SlotDefinition:
        if 'items' in obj:
            items_obj = obj['items']
            uniqueItems = obj.get('uniqueItems', False)
            if isinstance(items_obj, str):
                # found in DOSDP: TODO: check
                items_obj = {'type': 'string'}
            if 'properties' in items_obj:
                slot = SlotDefinition(name)
                slot.range = self.translate_object(items_obj)
            else:
                slot = self.translate_property(items_obj, name)
            if 'description' in items_obj:
                slot.description = items_obj['description']
            slot.multivalued = True
            return slot
        elif 'properties' in obj:
            c = ClassDefinition(f'{name}Class')
            self.translate_properties(obj, c)
            self.schema.classes[c.name] = c
            slot = SlotDefinition(name)
            slot.range = c.name
            slot.multivalued = True
            return slot
        else:
            logging.error(f'NOT HANDLED: {obj} in array context')
            return None

    def translate_ref(self, obj: dict) -> ClassDefinitionName:
        return ClassDefinitionName(self.get_id(obj))

    def translate_oneOf(self, oneOfList: List) -> Optional[ClassDefinition]:
        if all('$ref' in x for x in oneOfList):
            cns = [self.translate_ref(x) for x in oneOfList]
            n = '_'.join(cns)
            c = ClassDefinition(n, union_of=cns)
            return c
        else:
            logging.warning(f'Cannot yet handle oneOfs without refs: {oneOfList}')
            return None

    def translate_property(self, obj: Dict, name: str) -> SlotDefinition:
        #print(f'Translating property {name}: {obj}')
        if name is None:
            raise ValueError(f'Name not set for {obj}')
        schema = self.schema
        aliases = []
        if name in RESERVED:
            aliases.append(name)
            name = f'_{name}'
        s = SlotDefinition(name)
        t = obj.get('type', None)
        s.description = obj.get('description', None)
        if s.description is not None:
            s.description = s.description.strip()
        default = obj.get('default', None)
        if '$ref' in obj:
            s.range = self.translate_ref(obj)
        elif t == 'array':
            s = self.translate_array(obj, name)
            if s is None:
                raise ValueError(f'Cannot translate array {name} {obj}')
        elif t == 'number':
            s.range = 'float'
        elif t == 'boolean':
            s.range = 'boolean'
        elif t == 'float':
            s.range = 'float'
        elif t == 'integer':
            s.range = 'integer'
            s.minimum_value = obj.get('minimum_value', None)
            s.maximum_value = obj.get('maximum_value', None)
        elif t == 'string':
            if 'enum' in obj:
                pvs = obj['enum']
                ename = f'{name}_options'
                schema.enums[ename] = EnumDefinition(name=ename, permissible_values=pvs)
                s.range = ename
        else:
            logging.error(f'Cannot translate type {t} in {obj}')
        if s.name is schema.slots:
            logging.warning(f'TODO: unify alternate slots')
        schema.slots[s.name] = s
        return s

    def translate_object(self, obj: Dict, name: str = None) -> ClassDefinitionName:
        """
        Translates jsonschema obj of type object

        Generates a ClassDefinition, inserts it into the schema, and returns the name
        """
        schema = self.schema
        t = obj.get('type', None)
        desc = obj.get('description', None)
        additional = obj.get('additionalProperties', False) ## TODO
        if 'oneOf' in obj:
            unionCls = self.translate_oneOf(obj.get('oneOf'))
        else:
            unionCls = None
        allOf = obj.get('allOf', None)
        title = obj.get('title', None)
        required = obj.get('required', [])

        properties = obj.get('properties', {})
        discriminator = obj.get('discriminator', None)  # In OpenAPI but not JSON-Schema
        pkg = None
        if name is None:
            name = obj.get('title', None)
        else:
            name, pkg = self.split_name(name)
            if name is None:
                raise ValueError(f'Problem splitting name from package')
        if name is None:
            name = 'TODO'
        c = ClassDefinition(name, description=desc, from_schema=pkg)
        if unionCls:
            c.union_of = unionCls.union_of
        for k, v in properties.items():
            # TODO: reuse below
            #print(f'  PROP {k} = {v}')
            slot = self.translate_property(v, k)
            if slot.name in required:
                slot.required = True
            c.slots.append(slot.name)
            c.slot_usage[slot.name] = slot
        schema.classes[c.name] = c
        return c.name


    def translate_properties(self, obj: dict, parent_class: ClassDefinition):
        required = obj.get('required', [])
        for k, v in obj.get('properties',{}).items():
            slot = self.translate_property(v, k)
            if slot.name in required:
                slot.required = True
            parent_class.slots.append(slot.name)
            parent_class.slot_usage[slot.name] = slot


@click.command()
@click.argument('input')
@click.option('--name', '-n', required=True, help='ID of schema')
@click.option('--format', '-f', default='json', help='JSON Schema format - yaml or json')
@click.option('--output', '-o', help='output path')
def jsonschema2model(input, output, name, format, **args):
    """ Infer a model from JSON Schema """
    ie = JsonSchemaImportEngine()
    schema = ie.load(input, name=name, format=format)
    write_schema(schema, output)


if __name__ == '__main__':
    jsonschema2model()




